@@ -1,9 +1,10 @@
 use super::git::{
-    branch_setting, get_current_branch, get_toplevel, make_git_command, setting_exists,
+    get_current_branch, get_git_path, get_settings, get_toplevel, make_git_command, setting_exists,
+    short_branch, LocalBranchName, ReferenceSpec, SettingEntry,
 };
 use super::worktree::{
-    append_lines, base_tree, relative_path, stash_switch, Commit, CommitSpec, Commitish, GitStatus,
-    SomethingSpec, SwitchErr, Tree, Treeish,
+    append_lines, base_tree, relative_path, stash_switch, target_branch_setting, Commit, CommitErr,
+    CommitSpec, Commitish, GitStatus, SomethingSpec, SwitchErr, Tree, Treeish, WorktreeHead,
 };
 use enum_dispatch::enum_dispatch;
 use std::env;
@@ -87,7 +88,13 @@ impl ArgMaker for Diff {
         let mut cmd_args: Vec<String> = cmd_args.iter().map(|s| s.to_string()).collect();
         cmd_args.push(match &self.source {
             Some(source) => source.sha.to_owned(),
-            None => base_tree().get_tree_reference(),
+            None => match base_tree() {
+                Ok(tree) => tree.get_tree_reference(),
+                Err(err) => {
+                    eprintln!("{}", err);
+                    return Err(1);
+                }
+            },
         });
         if let Some(target) = &self.target {
             cmd_args.push(target.sha.to_owned());
@@ -122,6 +129,7 @@ impl ArgMaker for Log {
             cmd_args.push("--first-parent");
         }
         if self.patch {
+            cmd_args.push("-m");
             cmd_args.push("--patch");
         }
         if let Some(range) = &self.range {
@@ -138,12 +146,31 @@ impl ArgMaker for Log {
 
 #[derive(Debug, StructOpt)]
 pub struct Merge {
-    source: CommitSpec,
+    #[structopt(long, short)]
+    source: Option<CommitSpec>,
 }
 
 impl ArgMaker for Merge {
     fn make_args(self) -> Result<Vec<String>, i32> {
-        Ok(["merge", "--no-commit", "--no-ff", &self.source.spec]
+        let source = if let Some(source) = self.source {
+            source
+        } else {
+            match find_target() {
+                Ok(Some(target)) => {
+                    eprintln!("Using remembered value {:?}", short_branch(&target.spec));
+                    target
+                }
+                Ok(None) => {
+                    eprintln!("Source not supplied and no remembered source.");
+                    return Err(1);
+                }
+                Err(err) => {
+                    eprintln!("{}", err);
+                    return Err(1);
+                }
+            }
+        };
+        Ok(["merge", "--no-commit", "--no-ff", &source.spec]
             .iter()
             .map(|s| s.to_string())
             .collect())
@@ -153,7 +180,8 @@ impl ArgMaker for Merge {
 #[derive(Debug, StructOpt)]
 pub struct MergeDiff {
     /// The branch you would merge into.  (Though any commitish will work.)
-    target: CommitSpec,
+    #[structopt(long, short)]
+    target: Option<CommitSpec>,
     /// Use the meyers diff algorithm.  (Faster, can produce more confusing diffs.)
     #[structopt(long)]
     myers: bool,
@@ -163,17 +191,79 @@ pub struct MergeDiff {
     path: Vec<String>,
 }
 
+/**
+ * Find a commit spec to merge into.
+ * note: Errors could be caused by a failed status command instead of a failed parse.
+ */
+fn find_target() -> Result<Option<CommitSpec>, CommitErr> {
+    let branch_name: LocalBranchName = match GitStatus::new() {
+        Ok(GitStatus {
+            head: WorktreeHead::Attached { head, .. },
+            ..
+        }) => head.parse().unwrap(),
+        Err(err) => {
+            return Err(CommitErr::GitError(err));
+        }
+        _ => {
+            return Ok(None);
+        }
+    };
+    let mut remote = None;
+    let target_branch = {
+        let mut target_branch = None;
+        let prefix = branch_name.setting_name("");
+        let target_setting = target_branch_setting(&branch_name);
+        let remote_setting = branch_name.setting_name("remote");
+        for entry in get_settings(&prefix, &["oaf-target-branch", "remote"]) {
+            if let SettingEntry::Valid { key, value } = entry {
+                if key == target_setting {
+                    target_branch = Some(value);
+                } else if key == remote_setting {
+                    remote = Some(value);
+                }
+            }
+        }
+        if let Some(target_branch) = target_branch {
+            if let Some(remote) = remote {
+                let local_branch = LocalBranchName {
+                    name: short_branch(&target_branch),
+                };
+                local_branch.with_repo(remote).full()
+            } else {
+                target_branch
+            }
+        } else {
+            return Ok(None);
+        }
+    };
+    Ok(Some(target_branch.parse()?))
+}
+
 impl ArgMaker for MergeDiff {
     fn make_args(self) -> Result<Vec<String>, i32> {
         if let Err(..) = Commit::from_str("HEAD") {
             eprintln!("Cannot merge-diff: no commits in HEAD.");
             return Err(1);
         }
+        let target = match self.target {
+            Some(target) => target,
+            None => match find_target() {
+                Ok(Some(target)) => {
+                    eprintln!("Using remembered value {:?}", short_branch(&target.spec));
+                    target
+                }
+                Ok(None) => {
+                    eprintln!("Target not supplied and no saved target.");
+                    return Err(1);
+                }
+                Err(err) => {
+                    eprintln!("{}", err);
+                    return Err(1);
+                }
+            },
+        };
         Diff {
-            source: Some(
-                self.target
-                    .find_merge_base(&CommitSpec::from_str("HEAD").unwrap()),
-            ),
+            source: Some(target.find_merge_base(&CommitSpec::from_str("HEAD").unwrap())),
             target: None,
             myers: self.myers,
             name_only: self.name_only,
@@ -185,7 +275,9 @@ impl ArgMaker for MergeDiff {
 
 #[derive(Debug, StructOpt)]
 pub struct Pull {
+    ///The Remote entry to pull from
     remote: Option<String>,
+    ///The branch to pull from
     source: Option<String>,
 }
 
@@ -216,11 +308,18 @@ impl ArgMaker for Restore {
     fn make_args(self) -> Result<Vec<String>, i32> {
         let source = if let Some(source) = self.source {
             source
-        } else if let Ok(source) = SomethingSpec::from_str("HEAD") {
-            source
         } else {
-            eprintln!("Cannot restore: no commits in HEAD.");
-            return Err(1);
+            match SomethingSpec::from_str("HEAD") {
+                Ok(source) => source,
+                Err(CommitErr::NoCommit { .. }) => {
+                    eprintln!("Cannot restore: no commits in HEAD.");
+                    return Err(1);
+                }
+                Err(CommitErr::GitError(err)) => {
+                    eprintln!("{}", err);
+                    return Err(1);
+                }
+            }
         };
         let source = source.get_treeish_spec();
         let cmd_args = vec!["checkout", &source];
@@ -290,7 +389,9 @@ pub enum NativeCommand {
     Checkout,
     /// Show the status of changed and unknown files in the working tree.
     Status,
-    /// Tell git to ignore a file
+    /// Tell git to ignore a file.
+    ///
+    /// This updates the top-level .gitignore, not any lower ones.
     Ignore,
 }
 #[derive(Debug, StructOpt)]
@@ -337,7 +438,13 @@ pub trait Runnable {
 impl Runnable for CommitCmd {
     fn run(self) -> i32 {
         if !self.no_strict {
-            let status = GitStatus::new();
+            let status = match GitStatus::new() {
+                Ok(status) => status,
+                Err(err) => {
+                    eprintln!("{}", err);
+                    return 1;
+                }
+            };
             let untracked = status.untracked_filenames();
             if !untracked.is_empty() {
                 eprintln!("Untracked files are present:");
@@ -366,19 +473,31 @@ pub struct Push {
 
 impl Runnable for Push {
     fn run(self) -> i32 {
-        let branch = get_current_branch();
+        let branch = match get_current_branch().parse::<LocalBranchName>() {
+            Ok(branch) => branch,
+            Err(unhandled) => {
+                eprintln!("Unhandled: {}", unhandled.name);
+                return 1;
+            }
+        };
         let mut args;
-        if setting_exists(&branch_setting(&branch, "remote")) {
-            if !setting_exists(&branch_setting(&branch, "merge")) {
+        if setting_exists(&branch.setting_name("remote")) {
+            if !setting_exists(&branch.setting_name("merge")) {
                 panic!("Branch in unsupported state");
             }
             args = vec!["push".to_string()];
         } else {
-            if let Ok(head) = Commit::from_str("HEAD") {
-                head
-            } else {
-                eprintln!("Cannot push: no commits in HEAD.");
-                return 1;
+            if let Err(err) = Commit::from_str("HEAD") {
+                match err {
+                    CommitErr::NoCommit { .. } => {
+                        eprintln!("Cannot push: no commits in HEAD.");
+                        return 1;
+                    }
+                    CommitErr::GitError(err) => {
+                        eprintln!("{}", err);
+                        return 1;
+                    }
+                }
             };
             args = ["push", "-u", "origin", "HEAD"]
                 .iter()
@@ -396,25 +515,33 @@ impl Runnable for Push {
 #[derive(Debug, StructOpt)]
 pub struct Switch {
     /// The branch to switch to.
-    branch: String,
+    branch: LocalBranchName,
     #[structopt(long, short)]
     create: bool,
 }
 
 impl Runnable for Switch {
     fn run(self) -> i32 {
+        if self.create && !self.branch.is_valid() {
+            eprintln!("'{}' is not a valid branch name.", self.branch.name);
+            return 1
+        }
         match stash_switch(&self.branch, self.create) {
             Ok(()) => 0,
             Err(SwitchErr::BranchInUse { path }) => {
-                println!("Branch {} is already in use at {}", self.branch, path);
+                println!("Branch {} is already in use at {}", self.branch.name, path);
                 1
             }
             Err(SwitchErr::AlreadyExists) => {
-                eprintln!("Branch {} already exists", self.branch);
+                eprintln!("Branch {} already exists", self.branch.name);
                 1
             }
             Err(SwitchErr::NotFound) => {
-                eprintln!("Branch {} not found", self.branch);
+                eprintln!("Branch {} not found", self.branch.name);
+                1
+            }
+            Err(SwitchErr::GitError(err)) => {
+                eprintln!("{}", err);
                 1
             }
         }
@@ -432,11 +559,9 @@ pub struct FakeMerge {
 
 impl Runnable for FakeMerge {
     fn run(self) -> i32 {
-        let head = if let Ok(head) = Commit::from_str("HEAD") {
-            head
-        } else {
-            eprintln!("Cannot fake-merge: no commits in HEAD.");
-            return 1;
+        let head = match head_for_squash() {
+            Ok(head) => head,
+            Err(exit_status) => return exit_status,
         };
         let message = if let Some(msg) = &self.message {
             &msg
@@ -461,13 +586,20 @@ pub struct SquashCommit {
     message: Option<String>,
 }
 
+fn head_for_squash() -> Result<Commit, i32> {
+    if let Ok(head) = Commit::from_str("HEAD") {
+        Ok(head)
+    } else {
+        eprintln!("Cannot squash commit: no commits in HEAD.");
+        Err(1)
+    }
+}
+
 impl Runnable for SquashCommit {
     fn run(self) -> i32 {
-        let head = if let Ok(head) = Commit::from_str("HEAD") {
-            head
-        } else {
-            eprintln!("Cannot fake-merge: no commits in HEAD.");
-            return 1;
+        let head = match head_for_squash() {
+            Ok(head) => head,
+            Err(exit_status) => return exit_status,
         };
         let parent = head.find_merge_base(&self.branch_point);
         let message = if let Some(msg) = &self.message {
@@ -486,9 +618,9 @@ impl Runnable for SquashCommit {
 #[derive(Debug, StructOpt)]
 pub struct Checkout {
     /// The branch to switch to.
-    branch_name: String,
+    _branch_name: String,
     #[structopt(long, short)]
-    branch: bool,
+    _branch: bool,
 }
 
 impl Runnable for Checkout {
@@ -505,10 +637,52 @@ pub struct Status {}
 
 impl Runnable for Status {
     fn run(self) -> i32 {
-        let gs = GitStatus::new();
+        let gs = match GitStatus::new() {
+            Ok(status) => status,
+            Err(err) => {
+                eprintln!("{}", err);
+                return 1;
+            }
+        };
+        match &gs.head {
+            WorktreeHead::Attached { head, upstream, .. } => {
+                println!("On branch {}", head);
+                if let Some(upstream) = upstream {
+                    if upstream.added == 0 && upstream.removed == 0 {
+                        println!("Your branch is up to date with '{}'.", upstream.name);
+                    } else if upstream.added == 0 {
+                        println!(
+                            "Your branch is behind '{}' by {} commit(s), and can be fast-forwarded.", upstream.name, upstream.removed);
+                    } else if upstream.removed == 0 {
+                        println!(
+                            "Your branch is ahead of '{}' by {} commit(s).",
+                            upstream.name, upstream.added
+                        );
+                        println!("  (use \"oaf push\" to publish your local commits");
+                    } else {
+                        println!("Your branch and '{}' have diverged,", upstream.name);
+                        println!(
+                            "and have {} and {} different commits each, respectively.",
+                            upstream.added, upstream.removed
+                        );
+                        println!(
+                            "  (use \"oaf merge {}\" to merge the remote branch into yours)",
+                            upstream.name
+                        );
+                    }
+                }
+            }
+            WorktreeHead::Detached(_) => {}
+        }
         let mut gs_iter = gs.iter();
         let cwd = env::current_dir().expect("Need cwd");
-        let top = get_toplevel();
+        let top = match get_toplevel() {
+            Ok(top) => top,
+            Err(err) => {
+                eprintln!("{}", err);
+                return 1;
+            }
+        };
         let top_rel = cwd.strip_prefix(top).unwrap();
         for se in gs_iter.fix_removals() {
             let out = se.format_entry(&top_rel);
@@ -520,25 +694,44 @@ impl Runnable for Status {
 
 #[derive(Debug, StructOpt)]
 pub struct Ignore {
+    /// Ignores the file in the local repository, instead of the worktree .gitignore.
+    #[structopt(long)]
+    local: bool,
+    /// The list of files to ignore
     files: Vec<String>,
 }
 
 /// Best-effort canonicalization.
 ///
 /// Canonicalizes the portions of the path that exist, ignores the rest.
+/// Does not traverse terminal symlinks.
 fn normpath(path: &Path) -> io::Result<PathBuf> {
-    for ancestor in path.ancestors() {
+    let mut abspath = std::env::current_dir().unwrap();
+    abspath.push(path);
+    for ancestor in abspath.ancestors().skip(1) {
         if let Ok(canonical) = ancestor.canonicalize() {
-            return Ok(canonical.join(path.strip_prefix(ancestor).unwrap()));
+            return Ok(canonical.join(abspath.strip_prefix(ancestor).unwrap()));
         }
     }
-    path.canonicalize()
+    Ok(abspath)
+}
+
+fn add_ignores(new_files: Vec<String>, ignore_file: &Path) {
+    let ignores = fs::read_to_string(&ignore_file).expect("Can't read ignores");
+    fs::write(ignore_file, append_lines(ignores, new_files)).expect("Can't write .gitignore");
 }
 
 impl Runnable for Ignore {
     fn run(self) -> i32 {
         let mut new_files = vec![];
-        let top = PathBuf::from(get_toplevel());
+        let top = PathBuf::from(match get_toplevel() {
+            Ok(top) => top,
+            Err(err) => {
+                eprintln!("{}", err);
+                return 1;
+            }
+        });
+        let top = top.canonicalize().unwrap();
         for file in &self.files {
             let path = normpath(&PathBuf::from(file)).unwrap();
             let mut relpath = relative_path(&top, path)
@@ -551,9 +744,27 @@ impl Runnable for Ignore {
             }
             new_files.push(relpath);
         }
-        let ignore_file = top.join(".gitignore");
-        let ignores = fs::read_to_string(&ignore_file).expect("Can't read ignores");
-        fs::write(ignore_file, append_lines(ignores, new_files)).expect("Can't write .gitignore");
+        let ignore_file = if self.local {
+            get_git_path("info/exclude")
+        } else {
+            top.join(".gitignore")
+        };
+        add_ignores(new_files, &ignore_file);
         0
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    #[test]
+    fn test_target_branch_setting() {
+        assert_eq!(
+            target_branch_setting(&LocalBranchName {
+                name: "my-branch".to_string()
+            }),
+            "branch.my-branch.oaf-target-branch"
+        );
+    }
+}
