@@ -6,19 +6,21 @@
 // option. This file may not be copied, modified, or distributed
 // except according to those terms.
 use super::git::{
-    create_stash, delete_ref, eval_rev_spec, full_branch, get_toplevel, git_switch,
-    make_git_command, output_to_string, run_git_command, set_head, upsert_ref,
+    create_stash, delete_ref, eval_rev_spec, get_toplevel, git_switch, make_git_command,
+    output_to_string, run_git_command, set_head, set_setting, upsert_ref, GitError,
+    LocalBranchName, ReferenceSpec, SettingLocation,
 };
 use enum_dispatch::enum_dispatch;
 use std::collections::HashMap;
-use std::ffi::OsStr;
+use std::ffi::{OsStr, OsString};
 use std::fmt;
 use std::io::prelude::*;
+use std::os::unix::ffi::OsStringExt;
 use std::path::{Path, PathBuf, StripPrefixError};
 use std::process::{Output, Stdio};
 use std::str::FromStr;
 
-#[derive(Debug, Clone, Copy)]
+#[derive(Debug, Clone, Copy, PartialEq)]
 pub enum EntryLocationStatus {
     Unmodified,
     Modified,
@@ -72,7 +74,7 @@ pub fn relative_path<T: AsRef<OsStr>, U: AsRef<OsStr>>(
     return Ok(PathBuf::from(to.strip_prefix(from)?));
 }
 
-#[derive(Debug, Copy, Clone)]
+#[derive(Debug, Copy, Clone, PartialEq)]
 pub struct StatusEntry<'a> {
     pub state: EntryState<'a>,
     pub filename: &'a str,
@@ -89,6 +91,7 @@ impl StatusEntry<'_> {
                 _ => " ",
             },
             EntryState::Renamed { .. } => "R",
+            EntryState::Unmerged { .. } => "C",
         };
         let disk_char = match self.state {
             EntryState::Untracked => "?",
@@ -108,6 +111,12 @@ impl StatusEntry<'_> {
                 EntryLocationStatus::Modified => "M",
                 _ => "$",
             },
+            EntryState::Unmerged { state } => match state {
+                UnmergedState::BothModified => "M",
+                UnmergedState::Added(Changer::Both) | UnmergedState::Added(Changer::Us) => "A",
+                UnmergedState::Deleted(Changer::Both) | UnmergedState::Deleted(Changer::Us) => "D",
+                UnmergedState::Deleted(Changer::Them) | UnmergedState::Added(Changer::Them) => " ",
+            },
         };
         let rename_str = if let EntryState::Renamed { old_filename, .. } = self.state {
             format!(
@@ -216,7 +225,20 @@ impl<'a> Iterator for StatusIter<'a> {
                         old_filename: self.raw_entries.next().unwrap(),
                     }
                 }
-                _ => continue,
+                "u " => {
+                    let result = EntryState::Unmerged {
+                        state: remain[..2].parse().unwrap(),
+                    };
+                    remain = &remain[159..];
+                    result
+                }
+                "# " => {
+                    continue;
+                }
+                _ => {
+                    eprintln!("Unhandled: {}", line);
+                    continue;
+                }
             };
             let filename = remain;
             return Some(StatusEntry {
@@ -228,7 +250,43 @@ impl<'a> Iterator for StatusIter<'a> {
     }
 }
 
-#[derive(Debug, Copy, Clone)]
+#[derive(Debug, Copy, Clone, PartialEq)]
+pub enum Changer {
+    Both,
+    Us,
+    Them,
+}
+
+#[derive(Debug, Copy, Clone, PartialEq)]
+pub enum UnmergedState {
+    Added(Changer),
+    BothModified,
+    Deleted(Changer),
+}
+
+#[derive(Debug)]
+pub enum UnmergedStateParseError {
+    UnhandledString,
+}
+
+impl FromStr for UnmergedState {
+    type Err = UnmergedStateParseError;
+
+    fn from_str(spec: &str) -> Result<Self, Self::Err> {
+        Ok(match spec {
+            "DD" => UnmergedState::Deleted(Changer::Both),
+            "AU" => UnmergedState::Added(Changer::Us),
+            "UD" => UnmergedState::Deleted(Changer::Them),
+            "UA" => UnmergedState::Added(Changer::Them),
+            "DU" => UnmergedState::Deleted(Changer::Us),
+            "AA" => UnmergedState::Added(Changer::Both),
+            "UU" => UnmergedState::BothModified,
+            _ => return Err(UnmergedStateParseError::UnhandledString),
+        })
+    }
+}
+
+#[derive(Debug, Copy, Clone, PartialEq)]
 pub enum EntryState<'a> {
     Untracked,
     Ignored,
@@ -241,11 +299,114 @@ pub enum EntryState<'a> {
         tree_status: EntryLocationStatus,
         old_filename: &'a str,
     },
+    Unmerged {
+        state: UnmergedState,
+    },
+}
+
+#[derive(Debug, PartialEq)]
+pub enum BranchCommit {
+    Initial,
+    Oid(String),
+}
+
+#[derive(Debug, PartialEq)]
+pub struct UpstreamInfo {
+    pub name: String,
+    pub added: u16,
+    pub removed: u16,
+}
+
+#[derive(PartialEq, Debug)]
+pub enum WorktreeHead {
+    Detached(String),
+    Attached {
+        commit: BranchCommit,
+        head: String,
+        upstream: Option<UpstreamInfo>,
+    },
+}
+
+impl UpstreamInfo {
+    fn factory<'a>(mut raw_entries: impl Iterator<Item = &'a str>) -> Option<Self> {
+        let name = if let Some(raw_upstream) = raw_entries.next() {
+            let segments: Vec<&str> = raw_upstream.split("# branch.upstream ").collect();
+            if segments.len() == 2 {
+                segments[1].to_string()
+            } else {
+                return None;
+            }
+        } else {
+            return None;
+        };
+        let segments: Vec<&str> = raw_entries.next().unwrap().split("# branch.ab ").collect();
+        let data = if segments.len() == 2 {
+            segments[1]
+        } else {
+            panic!()
+        };
+        let (added, removed) = {
+            let mut ab = data.split(' ').map(|x| x[1..].parse::<u16>().unwrap());
+            (ab.next().unwrap(), ab.next().unwrap())
+        };
+        Some(UpstreamInfo {
+            name,
+            added,
+            removed,
+        })
+    }
+}
+
+pub fn make_worktree_head<'a>(mut raw_entries: impl Iterator<Item = &'a str>) -> WorktreeHead {
+    if let Some(raw_oid) = raw_entries.next() {
+        let oid = {
+            let segments: Vec<&str> = raw_oid.split("# branch.oid ").collect();
+            if segments.len() == 2 {
+                segments[1]
+            } else {
+                panic!()
+            }
+        };
+        let head = {
+            if let Some(raw_head) = raw_entries.next() {
+                let segments: Vec<&str> = raw_head.split("# branch.head ").collect();
+                if segments.len() == 2 {
+                    segments[1]
+                } else {
+                    panic!()
+                }
+            } else {
+                panic!()
+            }
+        };
+        if head == "(detached)" {
+            WorktreeHead::Detached(oid.to_string())
+        } else {
+            let upstream = UpstreamInfo::factory(raw_entries);
+            WorktreeHead::Attached {
+                commit: BranchCommit::Oid(oid.to_string()),
+                head: head.to_string(),
+                upstream,
+            }
+        }
+    } else {
+        WorktreeHead::Attached {
+            commit: BranchCommit::Initial,
+            head: "".to_string(),
+            upstream: Some(UpstreamInfo {
+                name: "".to_string(),
+                added: 0,
+                removed: 0,
+            }),
+        }
+    }
 }
 
 /// Represents `git status` output
+#[derive(Debug)]
 pub struct GitStatus {
     outstr: String,
+    pub head: WorktreeHead,
 }
 
 impl GitStatus {
@@ -260,11 +421,26 @@ impl GitStatus {
     }
 
     ///Return an [GitStatus] for the current directory
-    pub fn new() -> GitStatus {
-        let output =
-            run_git_command(&["status", "--porcelain=v2", "-z"]).expect("Couldn't list directory");
+    pub fn new() -> Result<GitStatus, GitError> {
+        let output = match run_git_command(&["status", "--porcelain=v2", "-z", "--branch"]) {
+            Err(output) => {
+                let stderr: OsString = OsStringExt::from_vec(output.stderr);
+                match GitError::from(stderr) {
+                    GitError::UnknownError(_) => {
+                        panic!("Couldn't list directory");
+                    }
+                    err => {
+                        return Err(err);
+                    }
+                }
+            }
+            Ok(output) => output,
+        };
         let outstr = output_to_string(&output);
-        GitStatus { outstr }
+        let info_iter = outstr.split_terminator('\0');
+        let head = make_worktree_head(info_iter);
+        let result = GitStatus { outstr, head };
+        Ok(result)
     }
 
     /** List untracked filenames
@@ -385,7 +561,7 @@ impl Commitish for Commit {
 #[derive(Debug)]
 pub struct CommitSpec {
     pub spec: String,
-    commit: Commit,
+    _commit: Commit,
 }
 
 #[enum_dispatch(Treeish)]
@@ -399,12 +575,24 @@ impl FromStr for SomethingSpec {
     type Err = CommitErr;
 
     fn from_str(spec: &str) -> Result<Self, CommitErr> {
-        let cmd = make_git_command(&["cat-file", "--batch-check"])
+        let mut cmd = make_git_command(&["cat-file", "--batch-check"])
             .stdin(Stdio::piped())
             .stdout(Stdio::piped())
+            .stderr(Stdio::piped())
             .spawn()
             .unwrap();
-        cmd.stdin.unwrap().write_all(spec.as_bytes()).unwrap();
+        cmd.stdin
+            .as_ref()
+            .unwrap()
+            .write_all(spec.as_bytes())
+            .unwrap();
+        if !cmd.wait().unwrap().success() {
+            let mut stderr_bytes = Vec::<u8>::new();
+            cmd.stderr.unwrap().read_to_end(&mut stderr_bytes).unwrap();
+            return Err(CommitErr::GitError(GitError::from(OsString::from_vec(
+                stderr_bytes,
+            ))));
+        }
         let mut result = String::new();
         cmd.stdout.unwrap().read_to_string(&mut result).unwrap();
         if result.ends_with("missing\n") {
@@ -418,7 +606,7 @@ impl FromStr for SomethingSpec {
         return Ok(match otype {
             "commit" => SomethingSpec::CommitSpec(CommitSpec {
                 spec: spec.to_string(),
-                commit: Commit {
+                _commit: Commit {
                     sha: oid.to_string(),
                 },
             }),
@@ -433,12 +621,14 @@ impl FromStr for SomethingSpec {
 #[derive(Debug)]
 pub enum CommitErr {
     NoCommit { spec: String },
+    GitError(GitError),
 }
 
 impl fmt::Display for CommitErr {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
             CommitErr::NoCommit { spec } => write!(f, "No commit found for \"{}\"", spec),
+            CommitErr::GitError(error) => error.fmt(f),
         }
     }
 }
@@ -449,7 +639,7 @@ impl FromStr for CommitSpec {
         let commit = Commit::from_str(spec)?;
         Ok(CommitSpec {
             spec: spec.to_string(),
-            commit,
+            _commit: commit,
         })
     }
 }
@@ -464,20 +654,24 @@ impl FromStr for Commit {
     type Err = CommitErr;
     fn from_str(spec: &str) -> std::result::Result<Self, <Self as FromStr>::Err> {
         match eval_rev_spec(spec) {
-            Err(..) => Err(CommitErr::NoCommit {
-                spec: spec.to_string(),
-            }),
+            Err(proc_output) => match GitError::from(OsStringExt::from_vec(proc_output.stderr)) {
+                GitError::UnknownError(_) => Err(CommitErr::NoCommit {
+                    spec: spec.to_string(),
+                }),
+                err => Err(CommitErr::GitError(err)),
+            },
             Ok(sha) => Ok(Commit { sha }),
         }
     }
 }
 
-pub fn base_tree() -> TreeSpec {
+pub fn base_tree() -> Result<TreeSpec, GitError> {
     let reference = match Commit::from_str("HEAD") {
         Ok(commit) => commit.get_tree_reference(),
-        Err(..) => "4b825dc642cb6eb9a060e54bf8d69288fbee4904".to_string(),
+        Err(CommitErr::NoCommit { .. }) => "4b825dc642cb6eb9a060e54bf8d69288fbee4904".to_string(),
+        Err(CommitErr::GitError(err)) => return Err(err),
     };
-    TreeSpec { reference }
+    Ok(TreeSpec { reference })
 }
 
 #[derive(Clone, Debug, PartialEq)]
@@ -539,7 +733,7 @@ pub fn list_worktree() -> Vec<WorktreeListEntry> {
 }
 
 pub fn create_wip_stash(wt: &WorktreeState) -> Option<String> {
-    let current_ref = make_wip_ref(wt);
+    let current_ref = WipReference::from_worktree_state(wt).full();
     match create_stash() {
         Some(oid) => {
             if let Err(..) = upsert_ref(&current_ref, &oid) {
@@ -557,12 +751,12 @@ pub fn create_wip_stash(wt: &WorktreeState) -> Option<String> {
 }
 
 pub fn apply_wip_stash(target_wt: &WorktreeState) -> bool {
-    let target_ref = make_wip_ref(target_wt);
-    match eval_rev_spec(&target_ref) {
+    let target_ref = WipReference::from_worktree_state(target_wt);
+    match target_ref.eval() {
         Err(..) => false,
         Ok(target_oid) => {
             run_git_command(&["stash", "apply", &target_oid]).unwrap();
-            delete_ref(&target_ref).unwrap();
+            delete_ref(&target_ref.full()).unwrap();
             true
         }
     }
@@ -581,11 +775,33 @@ pub fn make_wip_ref(wt: &WorktreeState) -> String {
     format!("refs/branch-wip/{}", splitted[1])
 }
 
-fn check_switch_branch(top: &str, branch: &str) -> Result<WorktreeListEntry, SwitchErr> {
+struct WipReference {
+    full_name: String,
+}
+
+impl WipReference {
+    pub fn from_worktree_state(wt: &WorktreeState) -> Self {
+        WipReference {
+            full_name: make_wip_ref(wt),
+        }
+    }
+}
+
+impl ReferenceSpec for WipReference {
+    fn full(&self) -> String {
+        self.full_name.to_string()
+    }
+}
+
+fn check_switch_branch(
+    top: &str,
+    branch: &LocalBranchName,
+) -> Result<WorktreeListEntry, SwitchErr> {
+    let top = PathBuf::from(top).canonicalize().unwrap();
     let mut self_wt = None;
-    let full_branch = full_branch(branch.to_string());
+    let full_branch = branch.full();
     for wt in list_worktree() {
-        if wt.path == top {
+        if PathBuf::from(&wt.path).canonicalize().unwrap() == top {
             self_wt = Some(wt);
             continue;
         }
@@ -605,14 +821,15 @@ pub enum SwitchErr {
     AlreadyExists,
     NotFound,
     BranchInUse { path: String },
+    GitError(GitError),
 }
 
 pub fn determine_switch_target(
-    branch: String,
+    branch: &LocalBranchName,
     create: bool,
     current_head: Option<&Commit>,
 ) -> Result<WorktreeState, SwitchErr> {
-    let full_branch = full_branch(branch.to_string());
+    let full_branch = branch.full();
     Ok(if let Ok(commit_id) = eval_rev_spec(&full_branch) {
         if create {
             return Err(SwitchErr::AlreadyExists);
@@ -632,12 +849,12 @@ pub fn determine_switch_target(
                 branch: full_branch,
             }
         }
-    } else if let Ok(commit_id) = eval_rev_spec(&format!("refs/remotes/origin/{}", branch)) {
+    } else if let Ok(commit_id) = branch.clone().with_repo("origin".to_string()).eval() {
         WorktreeState::CommittedBranch {
             branch: full_branch,
             head: Commit { sha: commit_id },
         }
-    } else if let Ok(commit_id) = eval_rev_spec(&branch) {
+    } else if let Ok(commit_id) = branch.eval() {
         WorktreeState::DetachedHead {
             head: Commit { sha: commit_id },
         }
@@ -646,15 +863,24 @@ pub fn determine_switch_target(
     })
 }
 
-pub fn stash_switch(branch: &str, create: bool) -> Result<(), SwitchErr> {
-    let top = get_toplevel();
-    let self_wt = check_switch_branch(&top, branch)?.state;
-    let self_head = match &self_wt {
-        WorktreeState::DetachedHead { head } => Some(head),
-        WorktreeState::CommittedBranch { head, .. } => Some(head),
-        WorktreeState::UncommittedBranch { .. } => None,
+pub fn target_branch_setting(branch: &LocalBranchName) -> String {
+    branch.setting_name("oaf-target-branch")
+}
+
+pub fn stash_switch(branch: &LocalBranchName, create: bool) -> Result<(), SwitchErr> {
+    let top = match get_toplevel() {
+        Ok(top) => top,
+        Err(err) => {
+            return Err(SwitchErr::GitError(err));
+        }
     };
-    let target_wt = determine_switch_target(branch.to_string(), create, self_head)?;
+    let self_wt = check_switch_branch(&top, branch)?.state;
+    let (self_head, old_branch) = match &self_wt {
+        WorktreeState::DetachedHead { head } => (Some(head), None),
+        WorktreeState::CommittedBranch { head, branch } => (Some(head), Some(branch)),
+        WorktreeState::UncommittedBranch { branch } => (None, Some(branch)),
+    };
+    let target_wt = determine_switch_target(branch, create, self_head)?;
     if create {
         eprintln!("Retaining any local changes.");
     } else if let Some(current_ref) = create_wip_stash(&self_wt) {
@@ -662,16 +888,20 @@ pub fn stash_switch(branch: &str, create: bool) -> Result<(), SwitchErr> {
     } else {
         eprintln!("No changes to stash");
     }
-    if let Err(..) = git_switch(branch, create, !create) {
-        panic!("Failed to switch to {}", branch);
+    if let Err(..) = git_switch(&branch.name, create, !create) {
+        panic!("Failed to switch to {}", branch.name);
     }
-    eprintln!("Switched to {}", branch);
+    eprintln!("Switched to {}", branch.name);
     if !create {
         if apply_wip_stash(&target_wt) {
-            eprintln!("Applied WIP changes for {}", branch);
+            eprintln!("Applied WIP changes for {}", branch.name);
         } else {
-            eprintln!("No WIP changes for {} to restore", branch);
+            eprintln!("No WIP changes for {} to restore", branch.name);
         }
+    } else if let Some(old_branch) = old_branch {
+        let name = target_branch_setting(branch);
+        set_setting(SettingLocation::Local, &name, old_branch)
+            .expect("Could not set target branch.");
     }
     Ok(())
 }
@@ -781,4 +1011,127 @@ mod tests {
         let contents2 = append_lines(contents, vec!["c".to_string()]);
         assert_eq!(contents2, "a\nb\nc\n");
     }
+
+    #[test]
+    fn test_parse_unmerged_state() {
+        assert_eq!(
+            "UU".parse::<UnmergedState>().unwrap(),
+            UnmergedState::BothModified
+        );
+        assert_eq!(
+            "AA".parse::<UnmergedState>().unwrap(),
+            UnmergedState::Added(Changer::Both)
+        );
+        assert_eq!(
+            "DD".parse::<UnmergedState>().unwrap(),
+            UnmergedState::Deleted(Changer::Both)
+        );
+        assert_eq!(
+            "AU".parse::<UnmergedState>().unwrap(),
+            UnmergedState::Added(Changer::Us)
+        );
+        assert_eq!(
+            "DU".parse::<UnmergedState>().unwrap(),
+            UnmergedState::Deleted(Changer::Us)
+        );
+        assert_eq!(
+            "UA".parse::<UnmergedState>().unwrap(),
+            UnmergedState::Added(Changer::Them)
+        );
+        assert_eq!(
+            "UD".parse::<UnmergedState>().unwrap(),
+            UnmergedState::Deleted(Changer::Them)
+        );
+    }
+
+    #[test]
+    fn test_conflict_entry() {
+        let mut iterator = StatusIter {
+            raw_entries: concat!(
+                "u UU N... 100755 100755 100755 100755 bcd098ed9e6b87c18c819847cab1cea07034635a",
+                " 9280a3393eeb5f48f43b5d47299f88308275624e",
+                " f1823404a82d732e4f6c33d7da256a563da8815a tools/btool.py"
+            )
+            .split_terminator('\0'),
+        };
+        assert_eq!(
+            iterator.next().unwrap(),
+            StatusEntry {
+                state: EntryState::Unmerged {
+                    state: UnmergedState::BothModified
+                },
+                filename: "tools/btool.py",
+            }
+        )
+    }
+
+    #[test]
+    fn test_make_worktree_head_detached() {
+        let info = make_worktree_head(
+            ["# branch.oid hello", "# branch.head (detached)"]
+                .iter()
+                .map(|x| *x),
+        );
+        if let WorktreeHead::Attached { commit, .. } = &info {
+            println!("{:?}", commit);
+        }
+        assert_eq!(info, WorktreeHead::Detached("hello".to_string(),));
+    }
+    #[test]
+    fn test_make_worktree_head_attached() {
+        let info = make_worktree_head(
+            ["# branch.oid hello", "# branch.head main"]
+                .iter()
+                .map(|x| *x),
+        );
+        assert_eq!(
+            info,
+            WorktreeHead::Attached {
+                commit: BranchCommit::Oid("hello".to_string()),
+                head: "main".to_string(),
+                upstream: None,
+            }
+        );
+    }
+    #[test]
+    fn test_make_worktree_head_attached_more() {
+        let info = make_worktree_head(
+            ["# branch.oid hello", "# branch.head main", "asdf"]
+                .iter()
+                .map(|x| *x),
+        );
+        assert_eq!(
+            info,
+            WorktreeHead::Attached {
+                commit: BranchCommit::Oid("hello".to_string()),
+                head: "main".to_string(),
+                upstream: None,
+            }
+        );
+    }
+    #[test]
+    fn test_make_worktree_head_upstream() {
+        let info = make_worktree_head(
+            [
+                "# branch.oid hello",
+                "# branch.head main",
+                "# branch.upstream origin/main",
+                "# branch.ab +25 -30",
+            ]
+            .iter()
+            .map(|x| *x),
+        );
+        assert_eq!(
+            info,
+            WorktreeHead::Attached {
+                commit: BranchCommit::Oid("hello".to_string()),
+                head: "main".to_string(),
+                upstream: Some(UpstreamInfo {
+                    name: "origin/main".to_string(),
+                    added: 25,
+                    removed: 30,
+                }),
+            }
+        );
+    }
 }
